#!/usr/bin/env python3
"""
TIN-X (Target Importance and Novelty Explorer) REST API client
https://www.newdrugtargets.org/
https://api.newdrugtargets.org/docs
https://api.newdrugtargets.org/targets/
"""
###
import sys,os,re,json,time,logging
import urllib,urllib.parse
#
from ..util import rest_utils
#
NCHUNK=100;
#
##############################################################################
def ListTargets(base_url, skip, nmax, fout):
  n_out=0; tags=None;
  url_next = (base_url+'/targets/?limit=%d&offset=%d'%(NCHUNK, skip))
  while True:
    rval = rest_utils.GetURL(url_next, parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    targets = rval["results"] if "results" in rval else []
    for target in targets:
      logging.debug(json.dumps(target, sort_keys=True, indent=2))
      if not tags:
        tags = list(target.keys())
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(target[tag]) if tag in target else '') for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    count = rval["count"] if "count" in rval else None
    if n_out%1000==0: logging.info("%d/%s done"%(n_out, count))
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
  logging.info("n_out: %d"%(n_out))

##############################################################################
def SearchTargets(base_url, terms, skip, nmax, fout):
  n_out=0; tags=None;
  for term in terms:
    url_next = (base_url+'/targets/?search=%s'%(urllib.parse.quote(term)))
    while True:
      rval = rest_utils.GetURL(url_next, parse_json=True)
      logging.debug(json.dumps(rval, sort_keys=True, indent=2))
      targets = rval["results"] if "results" in rval else []
      for target in targets:
        logging.debug(json.dumps(target, sort_keys=True, indent=2))
        if not tags:
          tags = list(target.keys())
          fout.write('\t'.join(tags)+'\n')
        vals = [(str(target[tag]) if tag in target else '') for tag in tags]
        fout.write('\t'.join(vals)+'\n')
        n_out+=1
        if nmax and n_out>=nmax: break
      if nmax and n_out>=nmax: break
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
  logging.info("n_out: %d"%(n_out))

##############################################################################
def ListDiseases(base_url, skip, nmax, fout):
  n_out=0; tags=None;
  url_next = (base_url+'/diseases/?limit=%d&offset=%d'%(NCHUNK, skip))
  while True:
    rval = rest_utils.GetURL(url_next, parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    diseases = rval["results"] if "results" in rval else []
    for disease in diseases:
      logging.debug(json.dumps(disease, sort_keys=True, indent=2))
      if not tags:
        tags = list(disease.keys())
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(disease[tag]) if tag in disease else '') for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    count = rval["count"] if "count" in rval else None
    if n_out%1000==0: logging.info("%d/%s done"%(n_out, count))
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
  logging.info("n_out: %d"%(n_out))

##############################################################################
def ListArticles(base_url, skip, nmax, fout):
  n_out=0; tags=None;
  url_next = (base_url+'/articles/?limit=%d&offset=%d'%(NCHUNK, skip))
  while True:
    rval = rest_utils.GetURL(url_next, parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    articles = rval["results"] if "results" in rval else []
    for article in articles:
      logging.debug(json.dumps(article, sort_keys=True, indent=2))
      if not tags:
        tags = list(article.keys())
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(article[tag]) if tag in article else '') for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    count = rval["count"] if "count" in rval else None
    if n_out%1000==0: logging.info("%d/%s done"%(n_out, count))
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
  logging.info("n_out: %d"%(n_out))

##############################################################################
def ListDTO(base_url, skip, nmax, fout):
  n_out=0; tags=None;
  url_next = (base_url+'/dto/?limit=%d&offset=%d'%(NCHUNK, skip))
  while True:
    rval = rest_utils.GetURL(url_next, parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    dtos = rval["results"] if "results" in rval else []
    for dto in dtos:
      logging.debug(json.dumps(dto, sort_keys=True, indent=2))
      if not tags:
        tags = list(dto.keys())
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(dto[tag]) if tag in dto else '') for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    count = rval["count"] if "count" in rval else None
    if n_out%1000==0: logging.info("%d/%s done"%(n_out, count))
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
  logging.info("n_out: %d"%(n_out))

##############################################################################
def GetDisease(base_url, ids, skip, nmax, fout):
  """IDs should be TIN-X disease IDs, e.g. 5391."""
  n_in=0; n_out=0; tags=None;
  for id_this in ids:
    n_in+=1
    if skip and skip>n_in: continue
    disease = rest_utils.GetURL((base_url+'/diseases/%s/'%id_this), parse_json=True)
    logging.debug(json.dumps(disease, sort_keys=True, indent=2))
    if not tags:
      tags = list(disease.keys())
      fout.write('\t'.join(tags)+'\n')
    vals = [(str(disease[tag]) if tag in disease else '') for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
    if nmax and n_in>=nmax: break
  logging.info("n_in: %d; n_out: %d"%(n_in, n_out))

##############################################################################
def GetDiseaseByDOId(base_url, ids, skip, nmax, fout):
  """IDs should be Disease Ontology IDs, e.g. DOID:9297."""
  n_in=0; n_out=0; tags=None;
  for id_this in ids:
    n_in+=1
    if skip and skip>n_in: continue
    rval = rest_utils.GetURL((base_url+'/diseases/?doid=%s'%id_this), parse_json=True)
    diseases = rval["results"] if "results" in rval else []
    for disease in diseases:
      logging.debug(json.dumps(disease, sort_keys=True, indent=2))
      if not tags:
        tags = list(disease.keys())
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(disease[tag]) if tag in disease else '') for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
    if nmax and n_in>=nmax: break
  logging.info("n_in: %d; n_out: %d"%(n_in, n_out))

##############################################################################
def GetTarget(base_url, ids, skip, nmax, fout):
  """IDs should be TIN-X target IDs, e.g. 10027."""
  n_in=0; n_out=0; tags=None;
  for id_this in ids:
    n_in+=1
    if skip and skip>n_in: continue
    target = rest_utils.GetURL((base_url+'/targets/%s/'), parse_json=True)
    logging.debug(json.dumps(target, sort_keys=True, indent=2))
    if not tags:
      tags = list(target.keys())
      fout.write('\t'.join(tags)+'\n')
    vals = [(str(target[tag]) if tag in target else '') for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
    if nmax and n_in>=nmax: break
  logging.info("n_in: %d; n_out: %d"%(n_in, n_out))

##############################################################################
def GetTargetByUniprot(base_url, ids, skip, nmax, fout):
  """IDs should be UniProt IDs, e.g. Q9H4B4."""
  n_in=0; n_out=0; tags=None;
  for id_this in ids:
    n_in+=1
    if skip and skip>n_in: continue
    rval = rest_utils.GetURL((base_url+'/targets/?uniprot=%s'), parse_json=True)
    targets = rval["results"] if "results" in rval else []
    for target in targets:
      logging.debug(json.dumps(target, sort_keys=True, indent=2))
      if not tags:
        tags = list(target.keys())
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(target[tag]) if tag in target else '') for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
    if nmax and n_in>=nmax: break
  logging.info("n_in: %d; n_out: %d"%(n_in, n_out))

##############################################################################
def GetTargetDiseases(base_url, ids, skip, nmax, fout):
  """IDs should be TIN-X target IDs, e.g. 10027."""
  n_in=0; n_out=0; tags=None;
  for id_this in ids:
    n_in+=1
    if skip and skip>n_in: continue
    while True:
      url_next = (base_url+'/targets/%s/diseases'%id_this)
      rval = rest_utils.GetURL(url_next, parse_json=True)
      diseases = rval["results"] if "results" in rval else []
      for disease in diseases:
        logging.debug(json.dumps(disease, sort_keys=True, indent=2))
        phenotype = disease["disease"] if "disease" in disease else {}
        if not tags:
          tags = list(disease.keys())
          tags.remove("disease")
          phenotype_tags = list(phenotype.keys())
          fout.write('\t'.join(["tinx_target_id"]+tags+phenotype_tags)+'\n')
        vals = [id_this]
        vals.extend([(str(disease[tag]) if tag in disease else '') for tag in tags])
        vals.extend([(str(phenotype[tag]) if tag in phenotype else '') for tag in phenotype_tags])
        fout.write('\t'.join(vals)+'\n')
        n_out+=1
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
    if nmax and n_in>=nmax: break
  logging.info("n_in: %d; n_out: %d"%(n_in, n_out))

##############################################################################
def GetDiseaseTargets(base_url, ids, skip, nmax, fout):
  """IDs should be TIN-X disease IDs, e.g. 5391."""
  n_in=0; n_out=0; tags=None;
  for id_this in ids:
    n_in+=1
    if skip and skip>n_in: continue
    while True:
      url_next = (base_url+'/diseases/%s/targets'%id_this)
      rval = rest_utils.GetURL(url_next, parse_json=True)
      targets = rval["results"] if "results" in rval else []
      for target in targets:
        logging.debug(json.dumps(target, sort_keys=True, indent=2))
        protein = target["target"] if "target" in target else {}
        if not tags:
          tags = list(target.keys())
          tags.remove("target")
          protein_tags = list(protein.keys())
          fout.write('\t'.join(["tinx_disease_id"]+tags+protein_tags)+'\n')
        vals = [id_this]
        vals.extend([(str(target[tag]) if tag in target else '') for tag in tags])
        vals.extend([(str(protein[tag]) if tag in protein else '') for tag in protein_tags])
        fout.write('\t'.join(vals)+'\n')
        n_out+=1
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
    if nmax and n_in>=nmax: break
  logging.info("n_in: %d; n_out: %d"%(n_in, n_out))

##############################################################################
def GetDiseaseTargetArticles(base_url, disease_ids, ids, skip, nmax, fout):
  """IDs should be TIN-X disease and target IDs and, e.g. 5391, 12203."""
  n_out=0; tags=None;
  for did in disease_ids:
    for tid in ids:
      url_next = (base_url+'/diseases/%s/targets/%s/articles'%(did, tid))
      while True:
        rval = rest_utils.GetURL(url_next, parse_json=True)
        articles = rval["results"] if "results" in rval else []
        for article in articles:
          logging.debug(json.dumps(article, sort_keys=True, indent=2))
        if not tags:
          tags = list(article.keys())
          fout.write('\t'.join(tags)+'\n')
        vals = [(str(article[tag]) if tag in article else '') for tag in tags]
        fout.write('\t'.join(vals)+'\n')
        n_out+=1
        url_next = rval["next"] if "next" in rval else None
        if not url_next: break
  logging.info("n_out: %d"%(n_out))

##############################################################################
def SearchDiseases(base_url, terms, skip, nmax, fout):
  """Search names; begins-with search logic."""
  n_out=0; tags=None;
  for term in terms:
    url_next = (base_url+'/diseases/?search=%s'%(urllib.parse.quote(term)))
    while True:
      rval = rest_utils.GetURL(url_next, parse_json=True)
      logging.debug(json.dumps(rval, sort_keys=True, indent=2))
      diseases = rval["results"] if "results" in rval else []
      for disease in diseases:
        logging.debug(json.dumps(disease, sort_keys=True, indent=2))
        if not tags:
          tags = list(disease.keys())
          fout.write('\t'.join(tags)+'\n')
        vals = [(str(disease[tag]) if tag in disease else '') for tag in tags]
        fout.write('\t'.join(vals)+'\n')
        n_out+=1
        if nmax and n_out>=nmax: break
      if nmax and n_out>=nmax: break
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
  logging.info("n_out: %d"%(n_out))

##############################################################################
def SearchTargets(base_url, terms, skip, nmax, fout):
  """Search names."""
  n_out=0; tags=None;
  for term in terms:
    url_next = (base_url+'/targets/?search=%s'%(urllib.parse.quote(term)))
    while True:
      rval = rest_utils.GetURL(url_next, parse_json=True)
      logging.debug(json.dumps(rval, sort_keys=True, indent=2))
      targets = rval["results"] if "results" in rval else []
      for target in targets:
        logging.debug(json.dumps(target, sort_keys=True, indent=2))
        if not tags:
          tags = list(target.keys())
          fout.write('\t'.join(tags)+'\n')
        vals = [(str(target[tag]) if tag in target else '') for tag in tags]
        fout.write('\t'.join(vals)+'\n')
        n_out+=1
        if nmax and n_out>=nmax: break
      if nmax and n_out>=nmax: break
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
  logging.info("n_out: %d"%(n_out))

##############################################################################
def SearchArticles(base_url, terms, skip, nmax, fout):
  n_out=0; tags=None;
  for term in terms:
    url_next = (base_url+'/articles/?search=%s'%(urllib.parse.quote(term)))
    while True:
      rval = rest_utils.GetURL(url_next, parse_json=True)
      logging.debug(json.dumps(rval, sort_keys=True, indent=2))
      articles = rval["results"] if "results" in rval else []
      for article in articles:
        logging.debug(json.dumps(article, sort_keys=True, indent=2))
        if not tags:
          tags = list(article.keys())
          fout.write('\t'.join(tags)+'\n')
        vals = [(str(article[tag]) if tag in article else '') for tag in tags]
        fout.write('\t'.join(vals)+'\n')
        n_out+=1
        if nmax and n_out>=nmax: break
      if nmax and n_out>=nmax: break
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
  logging.info("n_out: %d"%(n_out))




#!/usr/bin/env python3
"""
OpenFDA Adverse Events Reports REST API utility functions.

 Reports include multiple reactions with multiple drugs, i.e. many-to-many
 drug-reaction associations.  It is inherently unknown which drug or drug
 combination is explanatory.

 My API key: ==SEE $HOME/.fda.yaml==
 My email: ==SEE $HOME/.fda.yaml==
 
 Query params:
   search	(e.g. search=patient.drug.drugindication:"multiple+myeloma")
   count	(e.g. count=patient.drug.drugindication)
   limit	(e.g. limit=25)
   skip	(e.g. skip=100)
 
 https://api.fda.gov/drug/event.json?search=patient.drug.openfda.pharm_class_epc:"nonsteroidal+anti-inflammatory+drug"&count=patient.reaction.reactionmeddrapt.exact
 
 patient.reaction.reactionmeddrapt  - MedDRA term(s) for the reaction(s).
 
 patient.reaction.reactionmeddraversionpt - The MedDRA version that patient.reaction.reactionmeddrapt uses.
 
 patient.reaction.reactionoutcome - Outcome of the reaction/event at the time of last observation
	1 = recovered/resolved
	2 = recovering/resolving
	3 = not recovered/not resolved
	4 = recovered/resolved with sequelae
	5 = fatal
	6 = unknown
"""
import sys,os,re,time,json,logging
#
from ...util import rest_utils
#
REST_RETRY_NMAX=10
REST_RETRY_WAIT=5

DRUG_ID_FIELDS = ['unii','rxcui','nui','spl_id','product_ndc','package_ndc']
DRUG_NAME_FIELDS = ['generic_name','brand_name','substance_name']

##############################################################################
def GetCounts(base_url, tfrom, tto):
  txt=''
  try:
    rval=rest_utils.GetURL(base_url+'?search=receivedate:[%s+TO+%s]&count=receivedate'%(tfrom,tto))
    txt+=('%s\n'%(str(rval)))
  except Exception as e:
    logging.error('%s'%(e))
  return txt

#############################################################################
def Info(base_url):
  url=(base_url+'?search=(serious:1)&limit=1')
  try:
    rval=rest_utils.GetURL(url,parse_json=True)
  except Exception as e:
    logging.error('%s'%(e))
    return None
  meta = rval['meta']
  return meta

#############################################################################
def GetFields(base_url):
  '''Show fields where value is either (1) string, or (2) list of strings.
Sample first n records.  However no guarantee sampling contains all fields.
'''
  n=100
  url=(base_url+'?search=(serious:1)&limit=%d'%n)
  try:
    rval=rest_utils.GetURL(url,parse_json=True)
  except Exception as e:
    logging.error('%s'%(e))
    return None

  logging.info('NOTE: Fields from sampled records, N = %d'%n)
  fields=set()
  for result in rval['results']:
    fields |= GetFieldsIn(result,'')

  return sorted(list(fields))

#############################################################################
def GetFieldsIn(obj,path):
  fields=set()
  if type(obj) is str: return
  elif type(obj) is list:
    if len(obj)==0 or type(obj[0]) is str: return
    for obj2 in obj:
      fields|=GetFieldsIn(obj2,path)
  elif type(obj) is dict:
    for field in obj.keys():
      path_this=('%s%s%s'%(path,('.' if path else ''),field))
      if type(obj[field]) is str:
        fields.add(path_this)
      elif type(obj[field]) is list:
        if len(obj[field])>0:
          if type(obj[field][0]) is str:
            fields.add(path_this)
          else:
            for obj2 in obj[field]:
              fields|=(GetFieldsIn(obj2,path_this))
      else:
        fields|=GetFieldsIn(obj[field],path_this)
  return fields

#############################################################################
def DrugName(drug):
  name=''
  if 'openfda' in drug :
    for name_field in DRUG_NAME_FIELDS:
      if name_field in drug['openfda'] :
        if len(drug['openfda'][name_field])>0:
          name=drug['openfda'][name_field][0]
          break
    if not name:
      for id_field in DRUG_ID_FIELDS:
        if id_field in drug['openfda'] :
          if len(drug['openfda'][id_field])>0:
            name=('%s=%s'%(id_field,drug['openfda'][id_field][0]))
            break
  else:
    if 'medicinalproduct' in drug :
      name=drug['medicinalproduct']
  return name

#############################################################################
def DrugID(drug,id_field):
  if 'openfda' in drug :
    if id_field in drug['openfda'] :
      if len(drug['openfda'][id_field])>0:
        return drug['openfda'][id_field][0]
  return ''

#############################################################################
def Search(drug_cl, drug_ind, drug_unii, drug_ndc, drug_spl, tfrom, tto, serious, fatal, rawquery, nmax, fout, base_url, api_key):
  '''The API seems to disallow limit exceeding 100.  So we need to iterate using skip.'''
  rval=None
  url=(base_url+'?api_key=%s&search='%api_key)
  qrys=[]
  if drug_cl:	qrys.append('(patient.drug.openfda.pharm_class_epc:%s)'%(drug_cl.replace(' ','+')))
  if drug_ind:	qrys.append('(patient.drug.drugindication:%s)'%(drug_ind.replace(' ','+')))
  if drug_unii:	qrys.append('(patient.drug.openfda.unii:%s)'%(drug_unii))
  if drug_ndc:	qrys.append('(patient.drug.openfda.ndc:%s)'%(drug_ndc))
  if drug_spl:	qrys.append('(patient.drug.openfda.spl:%s)'%(drug_spl))
  if fatal:	qrys.append('(seriousnessdeath:1)')
  if serious:	qrys.append('(serious:1)')
  if tfrom:	qrys.append('(receivedate:[%s+TO+%s])'%(tfrom,tto))
  if rawquery:	qrys.append('(%s)'%rawquery.replace(' ','+'))
  if len(qrys)==0:
    logging.error('No query specified.')
    return rval

  url+=('+AND+'.join(qrys))

  drugnames_all = set()
  uniis_all = set()
  rxns_all = set()

  fout.write('ReportID,ReceiptDate,Drugname,UNII,ProductNDC,Seriousness,Event\n')

  ser_counts =  {s:0 for s in ('F','H','S','')}

  fromtime = time.localtime()
  totime = time.strptime('19000101','%Y%m%d')

  ndone=0; nchunk=100; n_report=0;
  while nmax==0 or ndone<nmax:
    if nmax>ndone: nchunk=min(nchunk,nmax-ndone)
    url_this=url+('&limit=%d'%(nchunk))+('&skip=%d'%ndone if ndone>0 else '')

    logging.debug('url="%s"'%url_this)
    try:
      rval=rest_utils.GetURL(url_this,parse_json=True)
    except Exception as e:
      logging.error('%s'%(e))
      break
    if not rval: break

    results = rval['results']
    logging.debug('results: %d'%len(results))

    for result in results:
      report_id = result['safetyreportid']
      if not 'drug' in result['patient'] :
        logging.error('report [ID=%s] missing drug[s].'%(report_id))
        continue
      n_report+=1
      reactions = result['patient']['reaction']
      recdate=result['receiptdate']
      rec_time = time.strptime(recdate,'%Y%m%d')
      fromtime = min(fromtime,rec_time)
      totime = max(totime,rec_time)

      drugs = result['patient']['drug']
      ser=('F' if 'seriousnessdeath' in result else ('H' if 'seriousnesshospitalization' in result else ('S' if 'serious' in result else '')))

      ser_counts[ser]+=1

      drugnames = set()
      uniis = set()
      rxns = set()

      for drug in drugs:
        drugname=DrugName(drug)
        unii=DrugID(drug,'unii')
        ndc=DrugID(drug,'product_ndc')
        drugnames.add(drugname)
        uniis.add(unii)

        for reaction in reactions:
          rxn=reaction['reactionmeddrapt']
          fout.write('%s,%s,"%s","%s","%s","%s","%s"\n'%(report_id,recdate,drugname,unii,ndc,ser,rxn))
          rxns.add(rxn)

      logging.debug('%d. Report: %s [%s] seriousness: %s'%(n_report,report_id,recdate,ser))
      logging.debug('\treactions: %s'%(', '.join(list(rxns))))
      logging.debug('\tdrugs: %s'%(', '.join(list(drugnames))))

      uniis_all |= uniis
      drugnames_all |= drugnames
      rxns_all |= rxns

    logging.debug(json.dumps(rval,indent=2))
    ndone+=nchunk
    if nmax>0 and ndone>=nmax: break

  logging.info('reports: %d ; drugs: %d ; reactions: %d'%(n_report,len(uniis_all),len(rxns_all)))
  logging.info('report seriousness: %s ; total: %d'%(str(ser_counts),sum(ser_counts.values())))
  logging.info('report date range: (%s - %s)'%(time.strftime('%d %b %Y',fromtime),time.strftime('%d %b %Y',totime)))

  return



#!/usr/bin/env python3
"""
https://mor.nlm.nih.gov/download/rxnav/RxNormAPIs.html
https://www.nlm.nih.gov/research/umls/rxnorm/docs/
"""
###
import sys,os,re,json,logging,urllib.parse

from ..util import rest_utils

#############################################################################
def List_IDTypes(base_url, fout):
  n_out=0;
  rval = rest_utils.GetURL(base_url+'/idtypes.json', parse_json=True)
  fout.write('IDType\n')
  for idtype in rval['idTypeList']['idName']:
    fout.write('%s\n'%idtype)
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def List_SourceTypes(base_url, fout):
  n_out=0;
  rval = rest_utils.GetURL(base_url+'/sourcetypes.json', parse_json=True)
  for sourcetype in rval['sourceTypeList']['sourceName']:
    fout.write('%s\n'%sourcetype)
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def List_RelationTypes(base_url, fout):
  n_out=0;
  rval = rest_utils.GetURL(base_url+'/relatypes.json', parse_json=True)
  for relatype in rval['relationTypeList']['relationType']:
    fout.write('%s\n'%relatype)
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def List_TermTypes(base_url, fout):
  n_out=0;
  rval = rest_utils.GetURL(base_url+'/termtypes.json', parse_json=True)
  for termtype in rval['termTypeList']['termType']:
    fout.write('%s\n'%termtype)
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def List_PropNames(base_url, fout):
  n_out=0;
  rval = rest_utils.GetURL(base_url+'/propnames.json', parse_json=True)
  for propname in rval['propNameList']['propName']:
    fout.write('%s\n'%propname)
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def List_PropCategories(base_url, fout):
  n_out=0;
  rval = rest_utils.GetURL(base_url+'/propCategories.json', parse_json=True)
  for propcat in rval['propCategoryList']['propCategory']:
    fout.write('%s\n'%propcat)
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def List_ClassTypes(base_url, fout):
  n_out=0;
  rval = rest_utils.GetURL(base_url+'/rxclass/classTypes.json', parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  for classtype in rval['classTypeList']['classTypeName']:
    fout.write('%s\n'%classtype)
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def List_Classes(base_url, class_types, fout):
  n_out=0;
  tags = ["classId", "classType", "className" ]
  url = (base_url+'/rxclass/allClasses.json')
  if class_types: url+=("?classTypes=%s"%urllib.parse.quote(' '.join(class_types)))
  rval = rest_utils.GetURL(url, parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  clss = rval["rxclassMinConceptList"]["rxclassMinConcept"] if "rxclassMinConceptList" in rval and "rxclassMinConcept" in rval["rxclassMinConceptList"] else []
  for cls in clss:
    vals = [cls[tag] if tag in cls else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_Name2RxCUI(base_url, names, fout):
  n_out=0;
  fout.write('Name\tRxCUI\n')
  for name in names:
    rval = rest_utils.GetURL(base_url+'/rxcui.json?name=%s'%urllib.parse.quote(name), parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    rxnormIds = rval["idGroup"]["rxnormId"] if "idGroup" in rval and "rxnormId" in rval["idGroup"] else []
    for rxnormId in rxnormIds:
      fout.write("%s\t%s\n"%(name, rxnormId))
      n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_Name(base_url, names, fout):
  n_out=0;
  tags = ["rxcui", "name", "synonym", "tty", "language", "suppress", "umlscui"]
  fout.write("rxcui\tname\tsynonym\ttty\tlanguage\tsuppress\tumlscui\n")
  for name in names:
    rval = rest_utils.GetURL(base_url+'/drugs.json?name=%s'%urllib.parse.quote(name), parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    conceptGroups = rval["drugGroup"]["conceptGroup"] if "drugGroup" in rval and "conceptGroup" in rval["drugGroup"] else []
    for cgroup in conceptGroups:
      cprops = cgroup["conceptProperties"] if "conceptProperties" in cgroup else []
      if not cprops: continue
      for cprop in cprops:
        vals = [cprop[tag] if tag in cprop else '' for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_ID2RxCUI(base_url, ids, idtype, fout):
  """For mapping external ID, for supported ID types, to RxNorm ID."""
  n_out=0;
  fout.write('%s\trxcui\n'%(idtype))
  for id_this in ids:
    rval = rest_utils.GetURL(base_url+'/rxcui.json?idtype=%s&id=%s'%(idtype, id_this), parse_json=True)
    for rxcui in rval['idGroup']['rxnormId']:
      fout.write('%s\t%s\n'%(id_this, rxcui))
      n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_RxCUI_Status(base_url, ids, fout):
  n_out=0;
  tags = ["rxcui", "name", "tty"]
  fout.write('\t'.join(tags)+"\tstatus\tremappedDate\n")
  for rxcui in ids:
    rval = rest_utils.GetURL(base_url+'/rxcui/%s/status.json'%rxcui,parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    if "rxcuiStatus" not in rval :
      logging.error("Bad record: %s"%str(rval))
      continue
    status = rval["rxcuiStatus"]["status"] if "status" in rval["rxcuiStatus"] else ""
    remappedDate = rval["rxcuiStatus"]["remappedDate"] if "remappedDate" in rval["rxcuiStatus"] else ""
    minConcepts = rval["rxcuiStatus"]["minConceptGroup"]["minConcept"] if "minConceptGroup" in rval["rxcuiStatus"] and "minConcept" in rval["rxcuiStatus"]["minConceptGroup"] else []
    for minConcept in minConcepts:
      vals = [minConcept[tag] if tag in minConcept else '' for tag in tags]
      fout.write('\t'.join(vals+[status, remappedDate])+"\n")
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_RxCUI_Properties(base_url, ids, fout):
  n_out=0; tags=None;
  for rxcui in ids:
    rval = rest_utils.GetURL(base_url+'/rxcui/%s/properties.json'%rxcui, parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    props = rval["properties"] if "properties" in rval else []
    if not tags:
      tags = props.keys()
      fout.write('\t'.join(tags)+'\n')
    vals = [props[tag] if tag in props else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_RxCUI_AllProperties(base_url, ids, fout):
  n_out=0; tags=None;
  for rxcui in ids:
    rval = rest_utils.GetURL(base_url+'/rxcui/%s/allProperties.json?prop=all'%rxcui, parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    props = rval["propConceptGroup"]["propConcept"] if "propConceptGroup" in rval and "propConcept" in rval["propConceptGroup"] else {}
    for prop in props:
      if not tags:
        tags = list(prop.keys())
        fout.write('\t'.join(['RxCui']+tags)+'\n')
      vals = [prop[tag] if tag in prop else '' for tag in tags]
      fout.write('\t'.join([rxcui]+vals)+'\n')
      n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_RxCUI_AllRelated(base_url, ids, fout):
  n_out=0; tags=None;
  for rxcui in ids:
    rval = rest_utils.GetURL(base_url+'/rxcui/%s/allrelated.json'%rxcui, parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    conceptGroups = rval["allRelatedGroup"]["conceptGroup"] if "allRelatedGroup" in rval and "conceptGroup" in rval["allRelatedGroup"] else []
    for cgroup in conceptGroups:
      props = cgroup["conceptProperties"] if "conceptProperties" in cgroup else []
      for prop in props:
        if not tags:
          tags = list(prop.keys())
          fout.write('\t'.join(['RxCui']+tags)+'\n')
        vals = [prop[tag] if tag in prop else '' for tag in tags]
        fout.write("\t".join([rxcui]+vals)+"\n")
        n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_RxCUI_NDCs(base_url, ids, fout):
  n_out=0;
  fout.write('rxcui\tndc\n')
  for rxcui in ids:
    rval = rest_utils.GetURL(base_url+'/rxcui/%s/ndcs.json'%rxcui, parse_json=True)
    ndcs = rval['ndcGroup']['ndcList']['ndc'] if 'ndcGroup' in rval and 'ndcList' in rval['ndcGroup'] and rval['ndcGroup']['ndcList'] and 'ndc' in rval['ndcGroup']['ndcList'] else []
    for ndc in ndcs:
      fout.write('%s\t%s\n'%(rxcui, ndc))
      n_out+=1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_RxCUI_Classes(base_url, ids, fout):
  n_out=0;
  tags = ["classId", "classType", "className" ]
  for rxcui in ids:
    rval = rest_utils.GetURL(base_url+'/rxcui/%s/classes.json'%rxcui, parse_json=True)
    logging.debug(json.dumps(rval, indent=4))

#############################################################################
def Get_Class_Members(base_url, class_id, rel_src, fout):
  n_out=0;
  url = (base_url+'/rxclass/%s/classMembers.json?classId=%s&relaSource=%s'%(class_id, rel_src))
  rval = rest_utils.GetURL(url, parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  #???
  logging.info("n_out: %d"%(n_out))

#############################################################################
def Get_SpellingSuggestions(base_url, name, fout):
  rval = rest_utils.GetURL(base_url+'/spellingsuggestions.json?name=%s'%urllib.parse.quote(name), parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  #???

#############################################################################
def RawQuery(base_url, rawquery, fout):
  rval = rest_utils.GetURL(base_url+rawquery, parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  #???

